
/*
  rzi.c
  rz assembly instruction generator

  Start: 8 Nov 2001
  Revised: 10 Nov 2001

  Program: Yodthong Rodkaew.

  Changes:
    Created:
       RZ.PRJ  TC++ 3.0 Project file
       RZI.C   RZ Compiler source C file
       RZI.H   RZ Compiler header file

    Modified:
       COMPILE.C  add some help comment (for me)
		  add code for RZI
       COMPILE.H  add code for RZI
       ICODE.C    add help comment
		  add code for RZI ( line-counter of source )
       LEX.C      change line counter info to RZI
       PARSE.C    bug and fix? add 3 line (see comment)
		  do not generate it again (fix PARSE.TXT first)

  Disclaimer: (10 Nov 2001)
   output file .S2 test with AS2 (S2-Assembler)
   but dose not test with (incoming) S2-Simulator
   I don't know it's work or not.

  11 Nov 2001:
   Tree version
    Created:
       RZT.C   tree manipulator

  change call sequence and fix some initial values
  P.Chongstitvatana  15 December 2001
  compact rzi.c  rzt.c and fix passing parameters
  abandon rzt.c  create z_xx
  fix unary minus at parse.c  item(){ ... tkMINUS }
  modify lex.c  dcl.c   doline() to track source line
  20 December 2001
  analyse number of reg used, buildfunctab()   29 Dec 2001
*/

#include "compile.h"
#include "rzi.h"

extern FILE *FO;	   /* compile.c */

typedef struct{ char str[10], len; } icEntry;
extern icEntry icProp[] ; /* iname.h */

int  stack_base = 1000;      /* program stack based */
int  data_base = 2000;       /* data base, global variable */
int  ar_base = 3000;		 /* base of activation record */
int  funcRef = 0;			 /* address of current function */

int  ip = 1;
int  doloc[MAXCS];
int  pass2 = 0;

/* ----- code gen stack ----- */

#define  ZMAX  100

typedef struct { int code; int p1; int p2; } zdata;
zdata zstack[ZMAX];
int zsp = 0;

/* --------- function table ----------- */

typedef struct{ int loc; int pv; int lv;
	int retq; int callq; int rmax;
	} functab_entry;

functab_entry functab[MAXFUNC];
int nfunc = 0;

/* -------- register allocation --------*/

int usereg = 0, rmax = 0;

void initR(){ usereg = 0;}

void initRX() { usereg = 0; rmax = 0;}

int newR(){
	usereg++;
	if(usereg > 28) {
		printf("use too many registers\n");
		usereg = 1;
	}
	if( usereg > rmax ) rmax = usereg;
	return usereg;
}

void freeR(){ if(usereg > 0) usereg--;}

/* ---------code gen evaluation stack ---------*/

void z_push(int code, int p1, int p2){
	if(zsp >= ZMAX) {
		fprintf(FO, "ztack overflow\n");
		return;
	}
	zsp++;
	zstack[zsp].code = code;
	zstack[zsp].p1 = p1;
	zstack[zsp].p2 = p2;
}

void z_pop(int *code, int *p1, int *p2){
	if(zsp < 0) {
		fprintf(FO,"zstack underflow\n");
		return;
	}
	*code = zstack[zsp].code;
	*p1 = zstack[zsp].p1;
	*p2 = zstack[zsp].p2;
	zsp--;
}

void z_init(){	zsp = 0;}

/* ---------------------------- */

int newsym(){
	static int num;
	num++;
	return num;
}

void prsrc(int line){	/* print source code */
	fprintf(FO, "   ;; ---- %d: %s\n", line ,getsrc(line));
}

/* create header file */
void rzheader()
{
/*
    fprintf(FO,";; --------------------------------- \n");
    fprintf(FO,";; \n");
    fprintf(FO,";; S2 assembly file v1.2 \n");
    fprintf(FO,";; \n");
    fprintf(FO,";; --------------------------------- \n");
    fprintf(FO,"\n");
    fprintf(FO,";; stack pointer register (sp) %s \n",sp_reg);
    fprintf(FO,";; base stack pointer register (bp) %s \n",bp_reg);
    fprintf(FO,";; save address return/call register (?) %s \n",addr_reg);
    fprintf(FO,";; zero register (zp) %s \n",zero_reg);
    fprintf(FO,"\n");
    fprintf(FO,";; general purpose register (1..4) r1 r2 r3 r4 \n");
    fprintf(FO,"\n");
    fprintf(FO,";; data based address %d \n",data_base);
    fprintf(FO,";; stack based address %d \n",stack_base);
    fprintf(FO,";; --------------------------------- \n");
    fprintf(FO,"\n\n");
*/
    fprintf(FO,".s\n");
    fprintf(FO," rz 0\n retv 28\n bp 29\n rads 30\n sp 31\n");
	fprintf(FO," print 1\n printch 2\n");
    fprintf(FO,".a 0\n");
    fprintf(FO,".c \n");
	fprintf(FO,"   ld sp #%d ;; initialise sp\n",stack_base);
	fprintf(FO,"   ld bp #%d ;; initialise bp\n",ar_base);
}

/* create end of file */
void rzend(){
    fprintf(FO,"\n.e \n");
}

/*------------------------------*/

/* get parameters into p[], ret no. of param */
int getparam(int ads, int *p){
	int len, i, n;
    len = icProp[CS[ads]].len;
	n = 1;
    for(i=ads+1; i<ads+len; i+=2) {
      p[n] = CS[i]*256 + CS[i+1];
	  n++;
	}
	return n-1;
}

/* if operand is lit 0, use r0 */
int genleaf(int code, int p1, int p2){
	int r1;
	switch( code ){
	case _Lval: r1=newR();fprintf(FO,"   sub  r%d bp #%d\n",r1,p1); break;
	case _Lvalg:r1=newR();fprintf(FO,"   ld   r%d #%d\n",r1,p1+data_base); break;
	case _Rval: r1=newR();fprintf(FO,"   ld   r%d @%d bp\n",r1,-p1); break;
	case _Rvalg:r1=newR();fprintf(FO,"   ld   r%d %d\n",r1,p1+data_base); break;
	case _Lit:
		if( p1 == 0 )
			r1 = 0;
		else {
			r1 = newR();
			fprintf(FO,"   ld   r%d #%d\n",r1,p1);
		}
		break;
	case _Reg:  r1 = p1; break;
	default:    fprintf(FO,"ERR: unknown opcode\n");
	}
	return r1;
}

int genopimm(int code, int reg, int p1){
	char a2[20];
	int dest;
	if(reg == 0) dest = newR();
	else dest = reg;
	if(p1 == 0) strcpy(a2,"rz");
	else sprintf(a2,"#%d",p1);
	switch(code){
	case _Add:   fprintf(FO,"   add  r%d r%d %s\n",dest,reg,a2); break;
	case _Sub:   fprintf(FO,"   sub  r%d r%d %s\n",dest,reg,a2); break;
	case _Mul:   fprintf(FO,"   mul  r%d r%d %s\n",dest,reg,a2); break;
	case _Div:   fprintf(FO,"   div  r%d r%d %s\n",dest,reg,a2); break;
	case _Index: fprintf(FO,"   add  r%d r%d %s\n",dest,reg,a2); break;
	case _And:   fprintf(FO,"   and  r%d r%d %s\n",dest,reg,a2); break;
	case _Or :   fprintf(FO,"   or   r%d r%d %s\n",dest,reg,a2); break;
	default :    fprintf(FO,"ERR: unknown opcode\n");
	}
	return dest;
}

/* it is a bit complicate dealing with r0 as
   if both operands are r0, a new reg must be allocated
   the lower number between r1, r2 is used as dest
   r0 will not be used as dest
  */
int getdest(int r1, int r2){
	int dest;
	if( r1 == 0 ){
		if( r2 == 0 ) dest = newR(); else dest = r2;
	}else {
		if( r2 == 0 ) dest = r1;
		else dest = r1 > r2 ? r2 : r1;
	}
	return dest;
}

/* must be careful about freeing a reg when r0 is involved */
int genop(int code, int r1, int r2){
	int dest;
	dest = getdest(r1,r2);
	switch(code){
	case _Add:   fprintf(FO,"   add  r%d r%d r%d\n",dest,r1,r2); break;
	case _Sub:   fprintf(FO,"   sub  r%d r%d r%d\n",dest,r1,r2); break;
	case _Mul:   fprintf(FO,"   mul  r%d r%d r%d\n",dest,r1,r2); break;
	case _Div:   fprintf(FO,"   div  r%d r%d r%d\n",dest,r1,r2); break;
	case _Index: fprintf(FO,"   add  r%d r%d r%d\n",dest,r1,r2); break;
	case _And:   fprintf(FO,"   and  r%d r%d r%d\n",dest,r1,r2); break;
	case _Or :   fprintf(FO,"   or   r%d r%d r%d\n",dest,r1,r2); break;
	default :    fprintf(FO,"ERR: unknown opcode\n");
	}
	if(r1 != 0) freeR();
	return dest;
}

int genop1(int code, int reg){
	switch(code){
	case _Fetch: fprintf(FO,"   ld   r%d @0 r%d\n",reg,reg); break;
	case _Minus: fprintf(FO,"   sub  r%d rz r%d\n",reg,reg); break;
	case _Not:   fprintf(FO,"   xor  r%d r%d #-1\n",reg,reg); break;
	default :    fprintf(FO,"ERR: unknown opcode\n");
	}
	return reg;
}

void z_set(){
	int op1,p11,p12,op2,p21,p22,a,v;
	z_pop(&op2,&p21,&p22);
	z_pop(&op1,&p11,&p12);
	v = genleaf(op2,p21,p22);
	switch(op1){
	case _Rval:
	case _Rvalg:
		a = genleaf(op1,p11,p12);
		fprintf(FO,"   st   @0 r%d r%d\n",a,v);
		break;
	case _Lval:  fprintf(FO,"   st   @%d bp r%d\n",-p11,v); break;
	case _Lvalg: fprintf(FO,"   st   @%d rz r%d\n",p11+data_base,v); break;
	case _Reg:   fprintf(FO,"   st   @0 r%d r%d\n",p11,v); break;
	default :    fprintf(FO,"ERR: unknown opcode\n");
	}
	initR();
	z_init();
}

void z_op2(int code){
	int op1,p11,p12,op2,p21,p22,r1,r2;
	z_pop(&op2,&p21,&p22);
	z_pop(&op1,&p11,&p12);
	r1 = genleaf(op1,p11,p12);
	if(op2 == _Lit){
		r1 = genopimm(code,r1,p21);
	}else {
		r2 = genleaf(op2,p21,p22);
		r1 = genop(code,r1,r2);
	}
	z_push(_Reg,r1,0);
}

void z_op1(int code) {
	int op,p1,p2,r1;
	z_pop(&op,&p1,&p2);
	r1 = genleaf(op,p1,p2);
	r1 = genop1(code,r1);
	z_push(_Reg,r1,0);
}

/* this is jump if false, so inverse condition */
void genjz(int cond, char *arg){
	switch(cond){
	case _Lt : fprintf(FO,"   jmp  ge %s\n",arg); break;
    case _Le : fprintf(FO,"   jmp  gt %s\n",arg); break;
    case _Eq : fprintf(FO,"   jmp  neq %s\n",arg); break;
    case _Ne : fprintf(FO,"   jmp  eq %s\n",arg); break;
    case _Ge : fprintf(FO,"   jmp  lt %s\n",arg); break;
    case _Gt : fprintf(FO,"   jmp  le %s\n",arg); break;
	default :  fprintf(FO,"ERR: unknown opcode\n");
	}
}

#define _RI    1	/* register immediate */
#define _RR    0	/* register register */

int gencond(int code, int r1, int r2, int mode){
	int p[4], dest, cond;
	char a2[20];
	dest = getdest(r1,r2);
	if(mode == _RI) sprintf(a2,"#%d",r2);
	else sprintf(a2,"r%d",r2);
	cond = code;
	code = CS[ip+1];		// read next code
	if( code == _Jz ){
		getparam(ip+1,p);
		fprintf(FO,"   sub  rz r%d %s\n",r1,a2);
		sprintf(a2,"label_%d",p[1]);
		genjz(cond,a2);
		initR();
	    fprintf(FO, "   ;; ---- Jz %d\n",p[1]);
		prsrc(CSLINE[ip+1]);
		ip += icProp[code].len;
	}else {
		fprintf(FO,"   sub  rz r%d %s\n",r1,a2);
		fprintf(FO,"   ld   r%d #0\n",dest);
		sprintf(a2,"skip_%d",newsym() );
		genjz(cond,a2);
		fprintf(FO,"   ld   r%d #1\n",dest);
		fprintf(FO,":%s\n",a2);
		z_push(_Reg,dest,0);
		if( r1 != 0) freeR();
	}
	return dest;
}

void z_cond(int code){
	int op1,p11,p12,op2,p21,p22,r1,r2;
	z_pop(&op2,&p21,&p22);
	z_pop(&op1,&p11,&p12);
	r1 = genleaf(op1,p11,p12);
	if(op2 == _Lit){
		r1 = gencond(code,r1,p21,_RI);
	}else {
		r2 = genleaf(op2,p21,p22);
		r1 = gencond(code,r1,r2,_RR);
	}
}

void z_jz(int ads){
	int op,p1,p2,r1;
	z_pop(&op,&p1,&p2);
	r1 = genleaf(op,p1,p2);
	if( r1 == 0 ){
		fprintf(FO,"   jmp  always label_%d\n",ads); // it is false, jump
	}else {
		fprintf(FO,"   or   rz r%d rz\n",r1);
		fprintf(FO,"   jmp  eq label_%d\n",ads);
		freeR();
	}
}


void buildFunctab(int loc){
	int p[5];
	nfunc++;
	if(nfunc >= MAXFUNC) seterror("too many func, functab overflow");
	getparam(loc,p);
	functab[nfunc].loc = loc;
	functab[nfunc].pv = p[1];
	functab[nfunc].lv = p[2];
	functab[nfunc].retq = 0;
	functab[nfunc].callq = 0;
	functab[nfunc].rmax = 0;
}

/* set functab[].rmax only in pass 2 */
void setrmax(){
	int i;
	if(pass2 == 1)
		if(CS[ip+1] == _Func || ip+1 >= CP){  // next is Func or end of cseg
			for(i=0; i<nfunc; i++)
				if(functab[i].loc == funcRef) break;
			functab[i].rmax = rmax;
		}
}

/* retrieve functab, ret index or 0 if not found */
int getfunc(int loc, int *pv, int *lv, int *retq, int *callq, int *rmax){
	int i;
	for(i=1; i<=nfunc; i++)
		if(functab[i].loc == loc) {
			*pv = functab[i].pv;
			*lv = functab[i].lv;
			*retq = functab[i].retq;
			*callq = functab[i].callq;
			*rmax = functab[i].rmax;
		}
	if( i <= nfunc ) return i;
	return 0;
}

void z_func(int loc){
	int pv, lv, retq, callq, rmax, r1, i;
	funcRef = loc;
	initRX();
	getfunc(loc,&pv,&lv,&retq,&callq,&rmax);
	fprintf(FO,"   add  bp bp #%d\n",pv+lv+1);
	if(callq)
		fprintf(FO,"   st   @0 bp rads\n");
	fprintf(FO,"   ;; ---- save regs\n");
	for(i=0; i<rmax; i++)
		fprintf(FO,"   st   @%d sp r%d\n",i+1,i+1);
	if( rmax > 0 ) fprintf(FO,"   add  sp sp #%d\n",rmax);
}

void z_ret0(int loc){
	int pv, lv, retq, callq, rmax, i;
	getfunc(loc,&pv,&lv,&retq,&callq,&rmax);
	if(callq)
		fprintf(FO,"   ld   rads @0 bp\n");
	fprintf(FO,"   sub  bp bp #%d\n",pv+lv+1);
	fprintf(FO,"   ;; ---- restore regs\n");
	for(i=0; i<rmax; i++)
		fprintf(FO,"   ld   r%d @%d sp\n",i+1,-(rmax-i-1));
	if( rmax > 0 ) fprintf(FO,"   sub  sp sp #%d\n",rmax);
	fprintf(FO,"   jr   rads\n");
	setrmax();
}

void z_ret1(int loc){
	int op, p1, p2, r1;
	z_pop(&op,&p1,&p2);
	r1 = p1;
	if(op != _Reg) r1 = genleaf(op,p1,p2);
	fprintf(FO, "   or   retv r%d rz\n",r1);
	freeR();
	z_ret0(loc);
}

void z_call(int loc){
	int pv, lv, retq, callq, rmax;
	int i, r1, op, p1, p2;
	getfunc(loc,&pv,&lv,&retq,&callq,&rmax);
	for(i=pv; i>0; i--) {
		z_pop(&op,&p1,&p2);
		r1 = genleaf(op,p1,p2);
		fprintf(FO,"   st   @%d bp r%d\n",i,r1);
		freeR();
	}
	fprintf(FO,"   jal  rads label_%d\n",loc);
	if(retq){
		r1 = newR();
		fprintf(FO,"   or   r%d retv rz\n",r1);
		z_push(_Reg,r1,0);
	}
}

void z_print(int code){
	int op, p1, p2, r1;
	z_pop(&op,&p1,&p2);
	r1 = genleaf(op,p1,p2);
	if(code == _Print)
		fprintf(FO,"   trap print r%d\n",r1);
	else
		fprintf(FO,"   trap printch r%d\n",r1);
	freeR();
}

void z_gen(int code, int p, int loc){
	switch( code ){
	case _Call : z_call(p); break;
	case _Jmp:   fprintf(FO, "   jmp always label_%d\n",p); break;
 	case _Jz :   z_jz(p); break;
	case _Func:  z_func(loc); break;
	case _Ret0:  z_ret0(funcRef); break;
	case _Ret1:  z_ret1(funcRef); break;
	case _Stop:	 fprintf(FO, "   trap  stop rz\n"); break;

    case _Lit:
    case _Lval :
    case _Lvalg:
    case _Rval:
    case _Rvalg: z_push(code,p,0); break;

    case _Fetch: z_op1(code); break;
    case _Set:   z_set(); break;

    case _And :
	case _Or  :
    case _Index:
    case _Add :
    case _Sub :
    case _Mul :
    case _Div :  z_op2(code); break;

    case _Minus :
    case _Not :  z_op1(code); break;

	case _Lt :
    case _Le :
    case _Eq :
    case _Ne :
    case _Ge :
    case _Gt :   z_cond(code); break;

    case _Print :
    case _Printch : z_print(code); break;

    default: fprintf(FO, " ;; %s not support!\n",icProp[code].str );
    }
}

void dumpfunctab(){
  int i;
  for(i=1; i<=nfunc; i++){
	  printf("loc %d pv %d lv %d retq %d callq %d rmax %d\n",
	    functab[i].loc,functab[i].pv,functab[i].lv,
		functab[i].retq,functab[i].callq,functab[i].rmax);
  }
}

void rzloop2()
{
  int code, param;	    /* mnemonic I-code and number of parameter */
  int i, p[5];

  /*----------------------------
     pass1 : check location reference and build functab[]
     set doloc[ addr ] =1 if someone reference the addr
  */
  for (i=0; i<CP; i++) doloc[i] = 0;
  ip = 1;
  while( ip < CP ) {
    code = CS[ip];
    param = getparam(ip,p);
    switch (code) {
    case _Jmp:
    case _Jz:   doloc[p[1]] = 1; break;
    case _Call: doloc[p[1]] = 1; functab[nfunc].callq = 1; break;
	case _Func: buildFunctab(ip); break;
	case _Ret0: functab[nfunc].retq = 0; break;
	case _Ret1: functab[nfunc].retq = 1; break;
    }
	ip += icProp[code].len;
  }

  /*----------------------------
   pass2 : generate S2 assembly code to set rmax
     base on I-code ( I-code to S2 translation )
  */

  pass2 = 1;
  ip = 1;
  while( ip < CP ) {
	code = CS[ip];
	param = getparam(ip,p);
	z_gen(code,p[1],ip);
	ip += icProp[code].len;
  }
  pass2 = 0;

// dumpfunctab();

 /*----------------------------
    pass 3   reset output file and generate with source listing
  */

  ip = 1;
  rewind(FO);
  rzheader();

  while( ip < CP ) {
    code = CS[ip];
	param = getparam(ip,p);

     /* create comment I-code , line no.: source */

    fprintf(FO, "   ;; ---- %s ", icProp[code].str );
	for(i=1; i<=param; i++)	fprintf(FO, "%d ", p[i]);
	fprintf(FO,"\n");
    if( code == _Set || code == _Jz || code == _Ret1) prsrc(CSLINE[ip]);

    /* create label if doloc[loc] ==1 */
    if (doloc[ip])
	    fprintf(FO, ":label_%d \n",ip);

	z_gen(code,p[1],ip);
	ip += icProp[code].len;
  }
  rzend();
}

/* end of file */
